@@ -22,6 +22,10 @@ _T = TypeVar('_T')
 
 
 class Config:
+  flax_use_flaxlib: bool
+  flax_array_ref: bool
+  flax_pytree_module: bool
+  flax_max_repr_depth: int | None
   # See https://google.github.io/pytype/faq.html.
   _HAS_DYNAMIC_ATTRIBUTES = True
 
@@ -62,6 +66,10 @@ class Config:
       raise LookupError(f'Unrecognized config option: {name}')
     self._values[name] = value
 
+  def __repr__(self):
+    values_repr = ', '.join(f'\n  {k}={v!r}' for k, v in self._values.items())
+    return f'Config({values_repr}\n)'
+
 
 config = Config()
 
@@ -117,6 +125,38 @@ def bool_flag(name: str, *, default: bool, help: str) -> FlagHolder[bool]:
   return fh
 
 
+def int_flag(name: str, *, default: int | None, help: str) -> FlagHolder[int]:
+  """Set up an integer flag.
+
+  Example::
+
+    num_foo = int_flag(
+        name='flax_num_foo',
+        default=42,
+        help='Number of foo.',
+    )
+
+  Now the ``FLAX_NUM_FOO`` shell environment variable can be used to
+  control the process-level value of the flag, in addition to using e.g.
+  ``config.update("flax_num_foo", 42)`` directly.
+
+  Args:
+    name: converted to lowercase to define the name of the flag. It is
+      converted to uppercase to define the corresponding shell environment
+      variable.
+    default: a default value for the flag.
+    help: used to populate the docstring of the returned flag holder object.
+
+  Returns:
+    A flag holder object for accessing the value of the flag.
+  """
+  name = name.lower()
+  config._add_option(name, static_int_env(name.upper(), default))
+  fh = FlagHolder[int](name, help)
+  setattr(Config, name, property(lambda _: fh.value, doc=help))
+  return fh
+
+
 def static_bool_env(varname: str, default: bool) -> bool:
   """Read an environment variable and interpret it as a boolean.
 
@@ -144,6 +184,27 @@ def static_bool_env(varname: str, default: bool) -> bool:
     )
 
 
+def static_int_env(varname: str, default: int | None) -> int | None:
+  """Read an environment variable and interpret it as an integer.
+
+  Args:
+    varname: the name of the variable
+    default: the default integer value
+  Returns:
+    integer return value derived from defaults and environment.
+  Raises: ValueError if the environment variable is not an integer.
+  """
+  val = os.getenv(varname)
+  if val is None:
+    return default
+  try:
+    return int(val)
+  except ValueError:
+    raise ValueError(
+      f'invalid integer value {val!r} for environment {varname!r}'
+    ) from None
+
+
 @contextmanager
 def temp_flip_flag(var_name: str, var_value: bool):
   """Context manager to temporarily flip feature flags for test functions.
@@ -162,9 +223,6 @@ def temp_flip_flag(var_name: str, var_value: bool):
 
 # Flax Global Configuration Variables:
 
-# Whether to use the lazy rng implementation.
-flax_lazy_rng = static_bool_env('FLAX_LAZY_RNG', True)
-
 flax_filter_frames = bool_flag(
   name='flax_filter_frames',
   default=True,
@@ -204,3 +262,25 @@ flax_fix_rng = bool_flag(
     ' PRNG keys.'
   ),
 )
+
+flax_use_flaxlib = bool_flag(
+  name='flax_use_flaxlib',
+  default=False,
+  help='Whether to use flaxlib for C++ acceleration.',
+)
+flax_array_ref = bool_flag(
+  name='flax_array_ref',
+  default=False,
+  help='Whether to use array refs.',
+)
+flax_pytree_module = bool_flag(
+  name='flax_pytree_module',
+  default=True,
+  help='Whether Modules are pytrees by default or not.',
+)
+
+flax_max_repr_depth = int_flag(
+  name='flax_max_repr_depth',
+  default=None,
+  help='Maximum depth of reprs for nested flax objects. Default is None (no limit).',
+)
\ No newline at end of file
