@@ -110,7 +110,10 @@ def _compute_stats(
   # but preserves double or complex floating points
   if force_float32_reductions:
     dtype = jnp.promote_types(dtype, jnp.float32)
-  x = jnp.asarray(x, dtype)
+  if isinstance(x, jax.Array):
+    x = x.astype(dtype)
+  else:
+    x = jnp.asarray(x, dtype)
   axes = _canonicalize_axes(x.ndim, axes)
 
   def maybe_distributed_mean(*xs, mask=None):
