@@ -19,4 +19,14 @@ from .wrappers import ToNNX as ToNNX
 from .wrappers import lazy_init as lazy_init
 from .wrappers import ToLinen as ToLinen
 from .wrappers import to_linen as to_linen
-from .variables import NNXMeta as NNXMeta
\ No newline at end of file
+from .variables import NNXMeta as NNXMeta
+from .variables import with_partitioning as with_partitioning
+from .module import Module as Module
+from .module import Scope as Scope
+from .module import AttrPriority as AttrPriority
+from .module import compact as compact
+from .module import current_context as current_context
+from .module import current_module as current_module
+from .interop import nnx_in_bridge_mdl as nnx_in_bridge_mdl
+from .interop import linen_in_bridge_mdl as linen_in_bridge_mdl
+from flax.nnx.nn import initializers as initializers
