@@ -15,21 +15,17 @@
 from __future__ import annotations
 
 import typing as tp
-from functools import partial
 
 import jax
 import jax.numpy as jnp
-import jax.tree_util as jtu
 
 from flax.nnx import (
   filterlib,
   graph,
 )
 from flax.nnx import variablelib as variableslib
-from flax.nnx.graph import GraphDef
-from flax.nnx.object import Object, ObjectMeta
-from flax.nnx.graph import GraphState, StateLeaf
-from flax.nnx.statelib import State
+from flax.nnx.pytreelib import Pytree, PytreeMeta
+from flax.nnx.graph import GraphState
 from flax.typing import Key, Path, PathParts
 
 A = tp.TypeVar('A')
@@ -44,13 +40,13 @@ tuple_reduce = lambda xs, x: xs + (x,)
 tuple_init = lambda: ()
 
 
-class ModuleMeta(ObjectMeta):
+class ModuleMeta(PytreeMeta):
   # we keep a trivial derived class just in case we need to
   # add more functionality in the future
   pass
 
 
-class Module(Object, metaclass=ModuleMeta):
+class Module(Pytree, metaclass=ModuleMeta):
   """Base class for all neural network modules.
 
   Layers and models should subclass this class.
@@ -82,13 +78,13 @@ class Module(Object, metaclass=ModuleMeta):
   """
 
   def sow(
-    self,
-    variable_type: tp.Type[variableslib.Variable[tp.Any]],
-    name: str,
-    value: A,
-    reduce_fn: tp.Callable[[B, A], B] = tuple_reduce,
-    init_fn: tp.Callable[[], B] = tuple_init,  # type: ignore
-  ) -> None:
+      self,
+      variable_type: type[variableslib.Variable[B]] | str,
+      name: str,
+      value: A,
+      reduce_fn: tp.Callable[[B, A], B] = tuple_reduce,
+      init_fn: tp.Callable[[], B] = tuple_init,  # type: ignore
+  ) -> bool:
     """``sow()`` can be used to collect intermediate values without
     the overhead of explicitly passing a container through each Module call.
     ``sow()`` stores a value in a new ``Module`` attribute, denoted by ``name``.
@@ -169,6 +165,11 @@ class Module(Object, metaclass=ModuleMeta):
         of ``init_fn`` together with the value to be stored. The default is an
         empty tuple.
     """
+    if isinstance(variable_type, str):
+      variable_type = variableslib.variable_type_from_name(
+          variable_type, allow_register=True
+      )
+
     if hasattr(self, name):
       variable = getattr(self, name)
       if not isinstance(variable, variableslib.Variable):
@@ -185,11 +186,15 @@ class Module(Object, metaclass=ModuleMeta):
       reduced_value = reduce_fn(init_fn(), value)
       setattr(self, name, variable_type(reduced_value))
 
+    return True
+
   def perturb(
-    self,
-    name: str,
-    value: tp.Any,
-    variable_type: tp.Type[variableslib.Variable[tp.Any]] = variableslib.Perturbation,
+      self,
+      name: str,
+      value: tp.Any,
+      variable_type: (
+          str | type[variableslib.Variable[tp.Any]]
+      ) = variableslib.Perturbation,
   ):
     """Add an zero-value variable ("perturbation") to the intermediate value.
 
@@ -246,10 +251,14 @@ class Module(Object, metaclass=ModuleMeta):
       variable_type: The :class:`Variable` type for the stored perturbation.
         Defaulted at :class:`nnx.Perturbation`.
     """
+    if isinstance(variable_type, str):
+      variable_type = variableslib.variable_type_from_name(
+          variable_type, allow_register=True
+      )
     if not hasattr(self, name):
       zeros = jax.tree.map(jnp.zeros_like, value)
       setattr(self, name, variable_type(zeros))
-    old_value = getattr(self, name)
+    old_value: variableslib.Variable[tp.Any] = getattr(self, name)
     if not isinstance(old_value, variable_type):
       raise ValueError(
         f"Expected '{name}' to be of type '{variable_type.__name__}', "
@@ -331,7 +340,9 @@ class Module(Object, metaclass=ModuleMeta):
       linear Linear
       submodule SubModule
     """
-    node_dict = graph.get_node_impl(self).node_dict(self)
+    node_impl = graph.get_node_impl(self)
+    assert node_impl is not None
+    node_dict = node_impl.node_dict(self)
     for key, value in node_dict.items():
       if isinstance(value, Module):
         yield key, value
@@ -466,41 +477,6 @@ class Module(Object, metaclass=ModuleMeta):
       raise_if_not_found=False,
     )
 
-  def __init_subclass__(cls, experimental_pytree: bool = False) -> None:
-    super().__init_subclass__()
-
-    if experimental_pytree:
-      jtu.register_pytree_with_keys(
-        cls,
-        partial(_module_flatten, with_keys=True),
-        _module_unflatten,  # type: ignore[arg-type]
-        flatten_func=partial(_module_flatten, with_keys=False),
-      )
-
-# -------------------------
-# Pytree Definition
-# -------------------------
-def _module_flatten(module: Module, *, with_keys: bool):
-  graphdef, state = graph.split(module)
-  key_values = sorted(state.raw_mapping.items())
-  keys = tuple(key for key, _ in key_values)
-
-  children: tuple[tp.Any, ...]
-  if with_keys:
-    children = tuple((jtu.DictKey(key), value) for key, value in key_values)
-  else:
-    children = tuple(value for _, value in key_values)
-
-  return children, (keys, graphdef)
-
-
-def _module_unflatten(
-  paths_moduledef: tuple[tuple[Path, ...], GraphDef[M]],
-  variables: tuple[StateLeaf, ...],
-) -> M:
-  paths, graphdef = paths_moduledef
-  return graph.merge(graphdef, State(zip(paths, variables)))
-
 
 def first_from(*args: tp.Optional[A], error_msg: str) -> A:
   """Return the first non-None argument.
