@@ -19,6 +19,7 @@ from __future__ import annotations
 import functools
 from typing import Any
 from collections.abc import Callable
+import math
 
 import jax
 import jax.numpy as jnp
@@ -28,7 +29,7 @@ from flax import nnx
 from flax.nnx import rnglib
 from flax.nnx.module import Module, first_from
 from flax.nnx.nn import initializers
-from flax.nnx.nn.dtypes import promote_dtype
+from flax.nnx.nn import dtypes
 from flax.nnx.nn.linear import (
   LinearGeneral,
   default_kernel_init,
@@ -36,6 +37,7 @@ from flax.nnx.nn.linear import (
 from flax.nnx.nn.normalization import LayerNorm
 from flax.typing import (
   Dtype,
+  PromoteDtypeFn,
   Shape,
   Initializer,
   PrecisionLike,
@@ -57,6 +59,7 @@ def dot_product_attention_weights(
   dtype: Dtype | None = None,
   precision: PrecisionLike = None,
   module: Module | None = None,
+  promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
 ):
   """Computes dot-product attention weights given query and key.
 
@@ -86,6 +89,9 @@ def dot_product_attention_weights(
     module: the Module that will sow the attention weights into the
       ``nnx.Intermediate`` collection. If ``module`` is None, the attention
       weights will not be sowed.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(query, key)`` and a ``dtype``
+      keyword argument, and return a tuple of arrays with the promoted dtype.
 
   Returns:
     Output of shape `[batch..., num_heads, q_length, kv_length]`.
@@ -148,6 +154,7 @@ def dot_product_attention(
   dtype: Dtype | None = None,
   precision: PrecisionLike = None,
   module: Module | None = None,
+  promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
 ):
   """Computes dot-product attention given query, key, and value.
 
@@ -155,6 +162,9 @@ def dot_product_attention(
   https://arxiv.org/abs/1706.03762. It calculates the attention weights given
   query and key and combines the values using the attention weights.
 
+  Will use the more optimized `jax.nn.dot_product_attention` if dropout is
+  not activated and `module=None`.
+
   .. note::
     ``query``, ``key``, ``value`` needn't have any batch dimensions.
 
@@ -182,6 +192,10 @@ def dot_product_attention(
     module: the Module that will sow the attention weights into the
       ``nnx.Intermediate`` collection. If ``module`` is None, the attention
       weights will not be sowed.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(query, key, value)`` and a
+      ``dtype`` keyword argument, and return a tuple of arrays with the promoted
+      dtype.
 
   Returns:
     Output of shape `[batch..., q_length, num_heads, v_depth_per_head]`.
@@ -197,6 +211,22 @@ def dot_product_attention(
   ), 'q, k, v num_heads must match.'
   assert key.shape[-3] == value.shape[-3], 'k, v lengths must match.'
 
+  # Criteria that invoke the more optimized dot product attention
+  if dropout_rate == 0.0 and module == None:
+    # make sure qkv batch are compressed to one dim
+    query_shape = query.shape
+    if len(query_shape) > 4:
+      def reshape_4d(x):
+        return jnp.reshape(x, (math.prod(x.shape[:-3]), *x.shape[-3:]))
+      query, key, value, bias, mask = jax.tree.map(
+        reshape_4d, (query, key, value, bias, mask))
+    if mask is not None:
+      mask = mask.astype(jnp.bool)
+    out = jax.nn.dot_product_attention(query, key, value, bias, mask)
+    if len(query_shape) > 4:
+      out = jnp.reshape(out, query_shape)
+    return out
+
   # compute attention weights
   attn_weights = dot_product_attention_weights(
     query,
@@ -249,7 +279,8 @@ class MultiHeadAttention(Module):
       should be divisible by the number of heads.
     in_features: int or tuple with number of input features.
     qkv_features: dimension of the key, query, and value.
-    out_features: dimension of the last projection
+    out_features: dimension of the last projection.
+    in_kv_features: number of input features for computing key and value.
     dtype: the dtype of the computation (default: infer from inputs and params)
     param_dtype: the dtype passed to parameter initializers (default: float32)
     broadcast_dropout: bool: use a broadcasted dropout along batch dims.
@@ -271,6 +302,10 @@ class MultiHeadAttention(Module):
     decode: whether to prepare and use an autoregressive cache.
     normalize_qk: should QK normalization be applied (arxiv.org/abs/2302.05442).
     rngs: rng key.
+    keep_rngs: whether to store the input rngs as attribute (i.e. `self.rngs = rngs`)
+      (default: True). If rngs is stored, we should split the module as
+      `graphdef, params, nondiff = nnx.split(module, nnx.Param, ...)` where `nondiff`
+      contains RNG object associated with stored `self.rngs`.
   """
 
   def __init__(
@@ -279,6 +314,7 @@ class MultiHeadAttention(Module):
     in_features: int,
     qkv_features: int | None = None,
     out_features: int | None = None,
+    in_kv_features: int | None = None,
     *,
     dtype: Dtype | None = None,
     param_dtype: Dtype = jnp.float32,
@@ -300,6 +336,7 @@ class MultiHeadAttention(Module):
     qkv_dot_general_cls: Any = None,
     out_dot_general_cls: Any = None,
     rngs: rnglib.Rngs,
+    keep_rngs: bool = True,
   ):
     self.num_heads = num_heads
     self.in_features = in_features
@@ -309,6 +346,9 @@ class MultiHeadAttention(Module):
     self.out_features = (
       out_features if out_features is not None else in_features
     )
+    self.in_kv_features = (
+      in_kv_features if in_kv_features is not None else in_features
+    )
     self.dtype = dtype
     self.param_dtype = param_dtype
     self.broadcast_dropout = broadcast_dropout
@@ -338,7 +378,6 @@ class MultiHeadAttention(Module):
 
     linear_general = functools.partial(
       LinearGeneral,
-      in_features=self.in_features,
       out_features=(self.num_heads, self.head_dim),
       dtype=self.dtype,
       param_dtype=self.param_dtype,
@@ -351,9 +390,9 @@ class MultiHeadAttention(Module):
     )
     # project inputs_q to multi-headed q/k/v
     # dimensions are then [batch..., length, n_heads, n_features_per_head]
-    self.query = linear_general(rngs=rngs)
-    self.key = linear_general(rngs=rngs)
-    self.value = linear_general(rngs=rngs)
+    self.query = linear_general(self.in_features, rngs=rngs)
+    self.key = linear_general(self.in_kv_features, rngs=rngs)
+    self.value = linear_general(self.in_kv_features, rngs=rngs)
 
     self.query_ln: LayerNorm | None
     self.key_ln: LayerNorm | None
@@ -392,7 +431,7 @@ class MultiHeadAttention(Module):
       dot_general_cls=self.out_dot_general_cls,
       rngs=rngs,
     )
-    self.rngs = rngs if dropout_rate > 0.0 else None
+    self.rngs = rngs.dropout.fork() if keep_rngs and dropout_rate > 0 else None
 
     self.cached_key: nnx.Cache[Array] | None = None
     self.cached_value: nnx.Cache[Array] | None = None
@@ -406,7 +445,7 @@ class MultiHeadAttention(Module):
     *,
     mask: Array | None = None,
     deterministic: bool | None = None,
-    rngs: rnglib.Rngs | None = None,
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
     sow_weights: bool = False,
     decode: bool | None = None,
   ):
@@ -445,6 +484,8 @@ class MultiHeadAttention(Module):
     """
     if rngs is None:
       rngs = self.rngs
+    elif isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.dropout
 
     if inputs_k is None:
       if inputs_v is not None:
@@ -507,14 +548,14 @@ class MultiHeadAttention(Module):
           % (expected_shape, query.shape)
         )
       # update key, value caches with our new 1d spatial slices
-      cur_index = self.cache_index.value
+      cur_index = self.cache_index[...]
       zero = jnp.array(0, dtype=lax.dtype(cur_index.dtype))
       indices = (zero,) * len(batch_dims) + (cur_index, zero, zero)
-      key = lax.dynamic_update_slice(self.cached_key.value, key, indices)
-      value = lax.dynamic_update_slice(self.cached_value.value, value, indices)
-      self.cached_key.value = key
-      self.cached_value.value = value
-      self.cache_index.value += 1
+      key = lax.dynamic_update_slice(self.cached_key[...], key, indices)
+      value = lax.dynamic_update_slice(self.cached_value[...], value, indices)
+      self.cached_key[...] = key
+      self.cached_value[...] = value
+      self.cache_index[...] += 1
       # causal mask for cached decoder self-attention:
       # our single query position should only attend to those key
       # positions that have already been generated and cached,
@@ -539,9 +580,10 @@ class MultiHeadAttention(Module):
       if not deterministic:
         if rngs is None:
           raise ValueError(
-            "'rngs' must be provided if 'dropout_rng' is not given."
+            "'rngs' must be provided to __call__ method if "
+            "MultiHeadAttention instance is defined with keep_rngs=False."
           )
-        dropout_rng = rngs.dropout()
+        dropout_rng = rngs()
       else:
         dropout_rng = None
     else:
