@@ -26,6 +26,7 @@ from flax import nnx
 from flax.nnx import rnglib, variablelib
 from flax.nnx.module import Module, first_from
 from flax.nnx.nn import dtypes, initializers
+from flax import nnx
 from flax.typing import (
   Dtype,
   Shape,
@@ -35,6 +36,8 @@ from flax.typing import (
   ConvGeneralDilatedT,
   PaddingLike,
   LaxPadding,
+  PromoteDtypeFn,
+  EinsumT,
 )
 
 Array = jax.Array
@@ -132,6 +135,10 @@ class LinearGeneral(Module):
     bias_init: initializer function for the bias.
     precision: numerical precision of the computation see ``jax.lax.Precision``
       for details.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(inputs, kernel, bias)``
+      and a ``dtype`` keyword argument, and return a tuple of arrays with the
+      promoted dtype.
     rngs: rng key.
   """
 
@@ -148,7 +155,7 @@ class LinearGeneral(Module):
     kernel_init: Initializer = default_kernel_init,
     bias_init: Initializer = default_bias_init,
     precision: PrecisionLike = None,
-    # Deprecated. Will be removed.
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
     dot_general: DotGeneralT | None = None,
     dot_general_cls: tp.Any = None,
     rngs: rnglib.Rngs,
@@ -165,6 +172,7 @@ class LinearGeneral(Module):
     self.precision = precision
     self.dot_general = dot_general
     self.dot_general_cls = dot_general_cls
+    self.promote_dtype = promote_dtype
 
     if len(self.in_features) != len(self.axis):
       raise ValueError(
@@ -257,7 +265,7 @@ class LinearGeneral(Module):
     batch_ind = tuple(range(n_batch_dims))
     contract_ind = tuple(range(n_batch_dims, n_axis + n_batch_dims))
 
-    inputs, kernel, bias = dtypes.promote_dtype(
+    inputs, kernel, bias = self.promote_dtype(
       (inputs, kernel, bias), dtype=self.dtype
     )
 
@@ -292,12 +300,10 @@ class Linear(Module):
     >>> layer = nnx.Linear(in_features=3, out_features=4, rngs=nnx.Rngs(0))
     >>> jax.tree.map(jnp.shape, nnx.state(layer))
     State({
-      'bias': VariableState(
-        type=Param,
+      'bias': Param(
         value=(4,)
       ),
-      'kernel': VariableState(
-        type=Param,
+      'kernel': Param(
         value=(3, 4)
       )
     })
@@ -313,6 +319,10 @@ class Linear(Module):
     kernel_init: initializer function for the weight matrix.
     bias_init: initializer function for the bias.
     dot_general: dot product function.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(inputs, kernel, bias)``
+      and a ``dtype`` keyword argument, and return a tuple of arrays with the
+      promoted dtype.
     rngs: rng key.
   """
 
@@ -328,6 +338,7 @@ class Linear(Module):
     kernel_init: Initializer = default_kernel_init,
     bias_init: Initializer = default_bias_init,
     dot_general: DotGeneralT = lax.dot_general,
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
     rngs: rnglib.Rngs,
   ):
     kernel_key = rngs.params()
@@ -350,6 +361,7 @@ class Linear(Module):
     self.kernel_init = kernel_init
     self.bias_init = bias_init
     self.dot_general = dot_general
+    self.promote_dtype = promote_dtype
 
   def __call__(self, inputs: Array) -> Array:
     """Applies a linear transformation to the inputs along the last dimension.
@@ -363,7 +375,7 @@ class Linear(Module):
     kernel = self.kernel.value
     bias = self.bias.value if self.bias is not None else None
 
-    inputs, kernel, bias = dtypes.promote_dtype(
+    inputs, kernel, bias = self.promote_dtype(
       (inputs, kernel, bias), dtype=self.dtype
     )
     y = self.dot_general(
@@ -409,6 +421,12 @@ class Einsum(Module):
       for details.
     kernel_init: initializer function for the weight matrix.
     bias_init: initializer function for the bias.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(inputs, kernel, bias)``
+      and a ``dtype`` keyword argument, and return a tuple of arrays with the
+      promoted dtype.
+    einsum_op: An injectable alternative of `jnp.einsum` to do the computation.
+      Should support same signature as `jnp.einsum`.
     rngs: rng key.
   """
 
@@ -423,6 +441,8 @@ class Einsum(Module):
     precision: PrecisionLike = None,
     kernel_init: Initializer = default_kernel_init,
     bias_init: Initializer = default_bias_init,
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
+    einsum_op: EinsumT = jnp.einsum,
     rngs: rnglib.Rngs,
   ):
     einsum_str = einsum_str.replace(' ', '')
@@ -446,6 +466,8 @@ class Einsum(Module):
     self.precision = precision
     self.kernel_init = kernel_init
     self.bias_init = bias_init
+    self.promote_dtype = promote_dtype
+    self.einsum_op = einsum_op
 
   def __call__(
     self, inputs: Array, einsum_str: tp.Optional[str] = None
@@ -472,7 +494,7 @@ class Einsum(Module):
     einsum_str = einsum_str.replace(' ', '')
     self._einsum_str_check(einsum_str)
 
-    inputs, kernel, bias = dtypes.promote_dtype(
+    inputs, kernel, bias = self.promote_dtype(
       (
         inputs,
         self.kernel.value,
@@ -481,7 +503,7 @@ class Einsum(Module):
       dtype=self.dtype,
     )
 
-    y = jnp.einsum(einsum_str, inputs, kernel, precision=self.precision)
+    y = self.einsum_op(einsum_str, inputs, kernel, precision=self.precision)
 
     if bias is not None:
       broadcasted_bias_shape = self._infer_broadcasted_bias_shape(
@@ -584,7 +606,8 @@ class Conv(Module):
     strides: an integer or a sequence of ``n`` integers, representing the
       inter-window strides (default: 1).
     padding: either the string ``'SAME'``, the string ``'VALID'``, the string
-      ``'CIRCULAR'`` (periodic boundary conditions), or a sequence of ``n``
+      ``'CIRCULAR'`` (periodic boundary conditions), the string `'REFLECT'`
+      (reflection across the padding boundary), or a sequence of ``n``
       ``(low, high)`` integer pairs that give the padding to apply before and after each
       spatial dimension. A single int is interpeted as applying the same padding
       in all dims and passign a single int in a sequence causes the same padding
@@ -609,6 +632,10 @@ class Conv(Module):
       for details.
     kernel_init: initializer for the convolutional kernel.
     bias_init: initializer for the bias.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(inputs, kernel, bias)``
+      and a ``dtype`` keyword argument, and return a tuple of arrays with the
+      promoted dtype.
     rngs: rng key.
   """
 
@@ -631,6 +658,7 @@ class Conv(Module):
     kernel_init: Initializer = default_kernel_init,
     bias_init: Initializer = default_bias_init,
     conv_general_dilated: ConvGeneralDilatedT = lax.conv_general_dilated,
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
     rngs: rnglib.Rngs,
   ):
     if isinstance(kernel_size, int):
@@ -670,6 +698,7 @@ class Conv(Module):
     self.kernel_init = kernel_init
     self.bias_init = bias_init
     self.conv_general_dilated = conv_general_dilated
+    self.promote_dtype = promote_dtype
 
   def __call__(self, inputs: Array) -> Array:
     """Applies a (potentially unshared) convolution to the inputs.
@@ -708,8 +737,7 @@ class Conv(Module):
     num_batch_dimensions = inputs.ndim - (len(kernel_size) + 1)
     if num_batch_dimensions != 1:
       input_batch_shape = inputs.shape[:num_batch_dimensions]
-      total_batch_size = int(np.prod(input_batch_shape))
-      flat_input_shape = (total_batch_size,) + inputs.shape[
+      flat_input_shape = (-1,) + inputs.shape[
         num_batch_dimensions:
       ]
       inputs = jnp.reshape(inputs, flat_input_shape)
@@ -720,17 +748,19 @@ class Conv(Module):
     kernel_dilation = maybe_broadcast(self.kernel_dilation)
 
     padding_lax = canonicalize_padding(self.padding, len(kernel_size))
-    if padding_lax == 'CIRCULAR':
+    if padding_lax in ('CIRCULAR', 'REFLECT'):
+      assert isinstance(padding_lax, str)
       kernel_size_dilated = [
         (k - 1) * d + 1 for k, d in zip(kernel_size, kernel_dilation)
       ]
-      zero_pad: tp.List[tuple[int, int]] = [(0, 0)]
+      zero_pad: list[tuple[int, int]] = [(0, 0)]
       pads = (
         zero_pad
         + [((k - 1) // 2, k // 2) for k in kernel_size_dilated]
         + [(0, 0)]
       )
-      inputs = jnp.pad(inputs, pads, mode='wrap')
+      padding_mode = {'CIRCULAR': 'wrap', 'REFLECT': 'reflect'}[padding_lax]
+      inputs = jnp.pad(inputs, pads, mode=padding_mode)
       padding_lax = 'VALID'
     elif padding_lax == 'CAUSAL':
       if len(kernel_size) != 1:
@@ -760,7 +790,7 @@ class Conv(Module):
 
     bias = self.bias.value if self.bias is not None else None
 
-    inputs, kernel, bias = dtypes.promote_dtype(
+    inputs, kernel, bias = self.promote_dtype(
       (inputs, kernel, bias), dtype=self.dtype
     )
 
@@ -857,6 +887,10 @@ class ConvTranspose(Module):
     bias_init: initializer for the bias.
     transpose_kernel: if ``True`` flips spatial axes and swaps the input/output
       channel axes of the kernel.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(inputs, kernel, bias)``
+      and a ``dtype`` keyword argument, and return a tuple of arrays with the
+      promoted dtype.
     rngs: rng key.
   """
 
@@ -877,6 +911,7 @@ class ConvTranspose(Module):
     kernel_init: Initializer = default_kernel_init,
     bias_init: Initializer = default_bias_init,
     transpose_kernel: bool = False,
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
     rngs: rnglib.Rngs,
   ):
     if isinstance(kernel_size, int):
@@ -898,6 +933,7 @@ class ConvTranspose(Module):
     self.kernel_init = kernel_init
     self.bias_init = bias_init
     self.transpose_kernel = transpose_kernel
+    self.promote_dtype = promote_dtype
 
     if self.transpose_kernel:
       kernel_shape = kernel_size + (self.out_features, in_features)
@@ -954,8 +990,7 @@ class ConvTranspose(Module):
     num_batch_dimensions = inputs.ndim - (len(kernel_size) + 1)
     if num_batch_dimensions != 1:
       input_batch_shape = inputs.shape[:num_batch_dimensions]
-      total_batch_size = int(np.prod(input_batch_shape))
-      flat_input_shape = (total_batch_size,) + inputs.shape[
+      flat_input_shape = (-1,) + inputs.shape[
         num_batch_dimensions:
       ]
       inputs = jnp.reshape(inputs, flat_input_shape)
@@ -982,7 +1017,7 @@ class ConvTranspose(Module):
 
     bias = self.bias.value if self.bias is not None else None
 
-    inputs, kernel, bias = dtypes.promote_dtype(
+    inputs, kernel, bias = self.promote_dtype(
       (inputs, kernel, bias), dtype=self.dtype
     )
 
@@ -1066,8 +1101,7 @@ class Embed(Module):
     >>> layer = nnx.Embed(num_embeddings=5, features=3, rngs=nnx.Rngs(0))
     >>> nnx.state(layer)
     State({
-      'embedding': VariableState( # 15 (60 B)
-        type=Param,
+      'embedding': Param( # 15 (60 B)
         value=Array([[ 0.57966787, -0.523274  , -0.43195742],
                [-0.676289  , -0.50300646,  0.33996582],
                [ 0.41796115, -0.59212935,  0.95934135],
@@ -1101,6 +1135,10 @@ class Embed(Module):
     dtype: the dtype of the embedding vectors (default: same as embedding).
     param_dtype: the dtype passed to parameter initializers (default: float32).
     embedding_init: embedding initializer.
+    promote_dtype: function to promote the dtype of the arrays to the desired
+      dtype. The function should accept a tuple of ``(embedding,)`` during ``__call__``
+      or ``(query, embedding)`` during ``attend``, and a ``dtype`` keyword argument,
+      and return a tuple of arrays with the promoted dtype.
     rngs: rng key.
   """
 
@@ -1112,6 +1150,7 @@ class Embed(Module):
     dtype: tp.Optional[Dtype] = None,
     param_dtype: Dtype = jnp.float32,
     embedding_init: Initializer = default_embed_init,
+    promote_dtype: PromoteDtypeFn = dtypes.promote_dtype,
     rngs: rnglib.Rngs,
   ):
     self.embedding = nnx.Param(
@@ -1123,6 +1162,7 @@ class Embed(Module):
     self.dtype = dtype or self.embedding.value.dtype
     self.param_dtype = param_dtype
     self.embedding_init = embedding_init
+    self.promote_dtype = promote_dtype
 
   def __call__(self, inputs: Array) -> Array:
     """Embeds the inputs along the last dimension.
@@ -1139,16 +1179,11 @@ class Embed(Module):
       raise ValueError('Input type must be an integer or unsigned integer.')
     # Use take because fancy indexing numpy arrays with JAX indices does not
     # work correctly.
-    (embedding,) = dtypes.promote_dtype(
+    (embedding,) = self.promote_dtype(
       (self.embedding.value,), dtype=self.dtype, inexact=False
     )
     if self.num_embeddings == 1:
-      return jnp.where(
-        jnp.broadcast_to(inputs[..., None], inputs.shape + (self.features,))
-        == 0,
-        embedding,
-        jnp.nan,
-      )
+      return jnp.broadcast_to(embedding, inputs.shape + (self.features,))
     return jnp.take(embedding, inputs, axis=0)
 
   def attend(self, query: Array) -> Array:
@@ -1164,7 +1199,7 @@ class Embed(Module):
       Commonly used for weight-sharing between embeddings and logit transform
       in NLP models.
     """
-    query, embedding = dtypes.promote_dtype(
+    query, embedding = self.promote_dtype(
       (query, self.embedding.value), dtype=self.dtype
     )
     return jnp.dot(query, embedding.T)
