@@ -50,7 +50,9 @@ class RNNCellBase(Module):
     """RNN cell base class."""
 
     def initialize_carry(
-        self, input_shape: tuple[int, ...], rngs: rnglib.Rngs | None = None
+      self,
+      input_shape: tuple[int, ...],
+      rngs: rnglib.Rngs | rnglib.RngStream | None = None,
     ) -> Carry:
         """Initialize the RNN cell carry.
 
@@ -91,7 +93,7 @@ def modified_orthogonal(key: Array, shape: Shape, dtype: Dtype = jnp.float32) ->
     return initializer(key, shape).astype(dtype)
 
 class LSTMCell(RNNCellBase):
-    r"""LSTM cell.
+  r"""LSTM cell.
 
   The mathematical definition of the cell is as follows
 
@@ -109,107 +111,123 @@ class LSTMCell(RNNCellBase):
   the memory.
   """
 
-    def __init__(
-        self,
-        in_features: int,
-        hidden_features: int,
-        *,
-        gate_fn: Callable[..., Any] = sigmoid,
-        activation_fn: Callable[..., Any] = tanh,
-        kernel_init: Initializer = default_kernel_init,
-        recurrent_kernel_init: Initializer = modified_orthogonal,
-        bias_init: Initializer = initializers.zeros_init(),
-        dtype: Dtype | None = None,
-        param_dtype: Dtype = jnp.float32,
-        carry_init: Initializer = initializers.zeros_init(),
-        rngs: rnglib.Rngs,
-    ):
-        self.in_features = in_features
-        self.hidden_features = hidden_features
-        self.gate_fn = gate_fn
-        self.activation_fn = activation_fn
-        self.kernel_init = kernel_init
-        self.recurrent_kernel_init = recurrent_kernel_init
-        self.bias_init = bias_init
-        self.dtype = dtype
-        self.param_dtype = param_dtype
-        self.carry_init = carry_init
-        self.rngs = rngs
+  def __init__(
+    self,
+    in_features: int,
+    hidden_features: int,
+    *,
+    gate_fn: Callable[..., Any] = sigmoid,
+    activation_fn: Callable[..., Any] = tanh,
+    kernel_init: Initializer = default_kernel_init,
+    recurrent_kernel_init: Initializer = modified_orthogonal,
+    bias_init: Initializer = initializers.zeros_init(),
+    dtype: Dtype | None = None,
+    param_dtype: Dtype = jnp.float32,
+    carry_init: Initializer = initializers.zeros_init(),
+    keep_rngs: bool = False,
+    rngs: rnglib.Rngs,
+  ):
+    self.in_features = in_features
+    self.hidden_features = hidden_features
+    self.gate_fn = gate_fn
+    self.activation_fn = activation_fn
+    self.kernel_init = kernel_init
+    self.recurrent_kernel_init = recurrent_kernel_init
+    self.bias_init = bias_init
+    self.dtype = dtype
+    self.param_dtype = param_dtype
+    self.carry_init = carry_init
+    self.rngs: rnglib.RngStream | None
+    if keep_rngs:
+      self.rngs = rngs.carry.fork()
+    else:
+      self.rngs = None
 
-        # input and recurrent layers are summed so only one needs a bias.
-        dense_i = partial(
-            Linear,
-            in_features=in_features,
-            out_features=hidden_features,
-            use_bias=False,
-            kernel_init=self.kernel_init,
-            dtype=self.dtype,
-            param_dtype=self.param_dtype,
-            rngs=rngs,
-        )
+    # input and recurrent layers are summed so only one needs a bias.
+    dense_i = partial(
+      Linear,
+      in_features=in_features,
+      out_features=hidden_features,
+      use_bias=False,
+      kernel_init=self.kernel_init,
+      dtype=self.dtype,
+      param_dtype=self.param_dtype,
+      rngs=rngs,
+    )
 
-        dense_h = partial(
-            Linear,
-            in_features=hidden_features,
-            out_features=hidden_features,
-            use_bias=True,
-            kernel_init=self.recurrent_kernel_init,
-            bias_init=self.bias_init,
-            dtype=self.dtype,
-            param_dtype=self.param_dtype,
-            rngs=rngs,
-        )
+    dense_h = partial(
+      Linear,
+      in_features=hidden_features,
+      out_features=hidden_features,
+      use_bias=True,
+      kernel_init=self.recurrent_kernel_init,
+      bias_init=self.bias_init,
+      dtype=self.dtype,
+      param_dtype=self.param_dtype,
+      rngs=rngs,
+    )
 
-        self.ii = dense_i()
-        self.if_ = dense_i()
-        self.ig = dense_i()
-        self.io = dense_i()
-        self.hi = dense_h()
-        self.hf = dense_h()
-        self.hg = dense_h()
-        self.ho = dense_h()
+    self.ii = dense_i()
+    self.if_ = dense_i()
+    self.ig = dense_i()
+    self.io = dense_i()
+    self.hi = dense_h()
+    self.hf = dense_h()
+    self.hg = dense_h()
+    self.ho = dense_h()
 
-    def __call__(self, carry: tuple[Array, Array], inputs: Array) -> tuple[tuple[Array, Array], Array]: # type: ignore[override]
-        r"""A long short-term memory (LSTM) cell.
+  def __call__(
+    self, carry: tuple[Array, Array], inputs: Array
+  ) -> tuple[tuple[Array, Array], Array]:  # type: ignore[override]
+    r"""A long short-term memory (LSTM) cell.
 
-        Args:
-          carry: the hidden state of the LSTM cell,
-            initialized using ``LSTMCell.initialize_carry``.
-          inputs: an ndarray with the input for the current time step.
-            All dimensions except the final are considered batch dimensions.
+    Args:
+      carry: the hidden state of the LSTM cell,
+        initialized using ``LSTMCell.initialize_carry``.
+      inputs: an ndarray with the input for the current time step.
+        All dimensions except the final are considered batch dimensions.
 
-        Returns:
-          A tuple with the new carry and the output.
-        """
-        c, h = carry
-        i = self.gate_fn(self.ii(inputs) + self.hi(h))
-        f = self.gate_fn(self.if_(inputs) + self.hf(h))
-        g = self.activation_fn(self.ig(inputs) + self.hg(h))
-        o = self.gate_fn(self.io(inputs) + self.ho(h))
-        new_c = f * c + i * g
-        new_h = o * self.activation_fn(new_c)
-        return (new_c, new_h), new_h
+    Returns:
+      A tuple with the new carry and the output.
+    """
+    c, h = carry
+    i = self.gate_fn(self.ii(inputs) + self.hi(h))
+    f = self.gate_fn(self.if_(inputs) + self.hf(h))
+    g = self.activation_fn(self.ig(inputs) + self.hg(h))
+    o = self.gate_fn(self.io(inputs) + self.ho(h))
+    new_c = f * c + i * g
+    new_h = o * self.activation_fn(new_c)
+    return (new_c, new_h), new_h
 
-    def initialize_carry(self, input_shape: tuple[int, ...], rngs: rnglib.Rngs | None = None) -> tuple[Array, Array]: # type: ignore[override]
-        """Initialize the RNN cell carry.
+  def initialize_carry(
+    self,
+    input_shape: tuple[int, ...],
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
+  ) -> tuple[Array, Array]:  # type: ignore[override]
+    """Initialize the RNN cell carry.
 
-        Args:
-          rng: random number generator passed to the init_fn.
-          input_shape: a tuple providing the shape of the input to the cell.
-        Returns:
-          An initialized carry for the given RNN cell.
-        """
-        batch_dims = input_shape[:-1]
-        if rngs is None:
-            rngs = self.rngs
-        mem_shape = batch_dims + (self.hidden_features,)
-        c = self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
-        h = self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
-        return (c, h)
+    Args:
+      rng: random number generator passed to the init_fn.
+      input_shape: a tuple providing the shape of the input to the cell.
+    Returns:
+      An initialized carry for the given RNN cell.
+    """
+    batch_dims = input_shape[:-1]
+    if rngs is None:
+      rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry
+    if rngs is None:
+      raise ValueError('RNGs must be provided to initialize the cell carry.')
 
-    @property
-    def num_feature_axes(self) -> int:
-        return 1
+    mem_shape = batch_dims + (self.hidden_features,)
+    c = self.carry_init(rngs(), mem_shape, self.param_dtype)
+    h = self.carry_init(rngs(), mem_shape, self.param_dtype)
+    return (c, h)
+
+  @property
+  def num_feature_axes(self) -> int:
+    return 1
 
 
 class OptimizedLSTMCell(RNNCellBase):
@@ -261,6 +279,7 @@ class OptimizedLSTMCell(RNNCellBase):
     dtype: Dtype | None = None,
     param_dtype: Dtype = jnp.float32,
     carry_init: Initializer = initializers.zeros_init(),
+    keep_rngs: bool = False,
     rngs: rnglib.Rngs,
   ):
     self.in_features = in_features
@@ -273,7 +292,11 @@ class OptimizedLSTMCell(RNNCellBase):
     self.dtype = dtype
     self.param_dtype = param_dtype
     self.carry_init = carry_init
-    self.rngs = rngs
+    self.rngs: rnglib.RngStream | None
+    if keep_rngs:
+      self.rngs = rngs.carry.fork()
+    else:
+      self.rngs = None
 
     # input and recurrent layers are summed so only one needs a bias.
     self.dense_i = Linear(
@@ -331,7 +354,9 @@ class OptimizedLSTMCell(RNNCellBase):
     return (new_c, new_h), new_h
 
   def initialize_carry(
-    self, input_shape: tuple[int, ...], rngs: rnglib.Rngs | None = None
+    self,
+    input_shape: tuple[int, ...],
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
   ) -> tuple[Array, Array]:  # type: ignore[override]
     """Initialize the RNN cell carry.
 
@@ -345,9 +370,13 @@ class OptimizedLSTMCell(RNNCellBase):
     batch_dims = input_shape[:-1]
     if rngs is None:
       rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry
+    if rngs is None:
+      raise ValueError('RNGs must be provided to initialize the cell carry.')
     mem_shape = batch_dims + (self.hidden_features,)
-    c = self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
-    h = self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
+    c = self.carry_init(rngs(), mem_shape, self.param_dtype)
+    h = self.carry_init(rngs(), mem_shape, self.param_dtype)
     return (c, h)
 
   @property
@@ -356,102 +385,116 @@ class OptimizedLSTMCell(RNNCellBase):
 
 
 class SimpleCell(RNNCellBase):
-    r"""Simple cell.
+  r"""Simple cell.
 
-    The mathematical definition of the cell is as follows
+  The mathematical definition of the cell is as follows
 
-    .. math::
+  .. math::
 
-        \begin{array}{ll}
-        h' = \tanh(W_i x + b_i + W_h h)
-        \end{array}
+      \begin{array}{ll}
+      h' = \tanh(W_i x + b_i + W_h h)
+      \end{array}
 
-    where x is the input and h is the output of the previous time step.
+  where x is the input and h is the output of the previous time step.
 
-    If `residual` is `True`,
+  If `residual` is `True`,
 
-    .. math::
+  .. math::
 
-        \begin{array}{ll}
-        h' = \tanh(W_i x + b_i + W_h h + h)
-        \end{array}
+      \begin{array}{ll}
+      h' = \tanh(W_i x + b_i + W_h h + h)
+      \end{array}
+  """
+
+  def __init__(
+    self,
+    in_features: int,
+    hidden_features: int,  # not inferred from carry for now
+    *,
+    dtype: Dtype = jnp.float32,
+    param_dtype: Dtype = jnp.float32,
+    carry_init: Initializer = initializers.zeros_init(),
+    residual: bool = False,
+    activation_fn: Callable[..., Any] = tanh,
+    kernel_init: Initializer = initializers.lecun_normal(),
+    recurrent_kernel_init: Initializer = initializers.orthogonal(),
+    bias_init: Initializer = initializers.zeros_init(),
+    keep_rngs: bool = False,
+    rngs: rnglib.Rngs,
+  ):
+    self.in_features = in_features
+    self.hidden_features = hidden_features
+    self.dtype = dtype
+    self.param_dtype = param_dtype
+    self.carry_init = carry_init
+    self.residual = residual
+    self.activation_fn = activation_fn
+    self.kernel_init = kernel_init
+    self.recurrent_kernel_init = recurrent_kernel_init
+    self.bias_init = bias_init
+    self.rngs: rnglib.RngStream | None
+    if keep_rngs:
+      self.rngs = rngs.carry.fork()
+    else:
+      self.rngs = None
+
+    # self.hidden_features = carry.shape[-1]
+    # input and recurrent layers are summed so only one needs a bias.
+    self.dense_h = Linear(
+      in_features=self.hidden_features,
+      out_features=self.hidden_features,
+      use_bias=False,
+      dtype=self.dtype,
+      param_dtype=self.param_dtype,
+      kernel_init=self.recurrent_kernel_init,
+      rngs=rngs,
+    )
+    self.dense_i = Linear(
+      in_features=self.in_features,
+      out_features=self.hidden_features,
+      use_bias=True,
+      dtype=self.dtype,
+      param_dtype=self.param_dtype,
+      kernel_init=self.kernel_init,
+      bias_init=self.bias_init,
+      rngs=rngs,
+    )
+
+  def __call__(self, carry: Array, inputs: Array) -> tuple[Array, Array]:  # type: ignore[override]
+    new_carry = self.dense_i(inputs) + self.dense_h(carry)
+    if self.residual:
+      new_carry += carry
+    new_carry = self.activation_fn(new_carry)
+    return new_carry, new_carry
+
+  def initialize_carry(
+    self,
+    input_shape: tuple[int, ...],
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
+  ) -> Array:  # type: ignore[override]
+    """Initialize the RNN cell carry.
+
+    Args:
+      rng: random number generator passed to the init_fn.
+      input_shape: a tuple providing the shape of the input to the cell.
+
+    Returns:
+      An initialized carry for the given RNN cell.
     """
+    if rngs is None:
+      rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry
+    if rngs is None:
+      raise ValueError('RNGs must be provided to initialize the cell carry.')
 
-    def __init__(
-        self,
-        in_features: int,
-        hidden_features: int,  # not inferred from carry for now
-        *,
-        dtype: Dtype = jnp.float32,
-        param_dtype: Dtype = jnp.float32,
-        carry_init: Initializer = initializers.zeros_init(),
-        residual: bool = False,
-        activation_fn: Callable[..., Any] = tanh,
-        kernel_init: Initializer = initializers.lecun_normal(),
-        recurrent_kernel_init: Initializer = initializers.orthogonal(),
-        bias_init: Initializer = initializers.zeros_init(),
-        rngs: rnglib.Rngs,
-    ):
-        self.in_features = in_features
-        self.hidden_features = hidden_features
-        self.dtype = dtype
-        self.param_dtype = param_dtype
-        self.carry_init = carry_init
-        self.residual = residual
-        self.activation_fn = activation_fn
-        self.kernel_init = kernel_init
-        self.recurrent_kernel_init = recurrent_kernel_init
-        self.bias_init = bias_init
-        self.rngs = rngs
+    batch_dims = input_shape[:-1]
+    mem_shape = batch_dims + (self.hidden_features,)
+    return self.carry_init(rngs(), mem_shape, self.param_dtype)
 
-        # self.hidden_features = carry.shape[-1]
-        # input and recurrent layers are summed so only one needs a bias.
-        self.dense_h = Linear(
-            in_features=self.hidden_features,
-            out_features=self.hidden_features,
-            use_bias=False,
-            dtype=self.dtype,
-            param_dtype=self.param_dtype,
-            kernel_init=self.recurrent_kernel_init,
-            rngs=rngs,
-        )
-        self.dense_i = Linear(
-            in_features=self.in_features,
-            out_features=self.hidden_features,
-            use_bias=True,
-            dtype=self.dtype,
-            param_dtype=self.param_dtype,
-            kernel_init=self.kernel_init,
-            bias_init=self.bias_init,
-            rngs=rngs,
-        )
-
-    def __call__(self, carry: Array, inputs: Array) -> tuple[Array, Array]: # type: ignore[override]
-        new_carry = self.dense_i(inputs) + self.dense_h(carry)
-        if self.residual:
-            new_carry += carry
-        new_carry = self.activation_fn(new_carry)
-        return new_carry, new_carry
-
-    def initialize_carry(self, input_shape: tuple[int, ...], rngs: rnglib.Rngs | None = None) -> Array: # type: ignore[override]
-        """Initialize the RNN cell carry.
-
-        Args:
-          rng: random number generator passed to the init_fn.
-          input_shape: a tuple providing the shape of the input to the cell.
-
-        Returns:
-          An initialized carry for the given RNN cell.
-        """
-        if rngs is None:
-            rngs = self.rngs
-        batch_dims = input_shape[:-1]
-        mem_shape = batch_dims + (self.hidden_features,)
-        return self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
-
-    @property
-    def num_feature_axes(self) -> int:
-        return 1
+  @property
+  def num_feature_axes(self) -> int:
+    return 1
 
 
 class GRUCell(RNNCellBase):
@@ -498,6 +541,7 @@ class GRUCell(RNNCellBase):
     dtype: Dtype | None = None,
     param_dtype: Dtype = jnp.float32,
     carry_init: Initializer = initializers.zeros_init(),
+    keep_rngs: bool = False,
     rngs: rnglib.Rngs,
   ):
     self.in_features = in_features
@@ -510,7 +554,11 @@ class GRUCell(RNNCellBase):
     self.dtype = dtype
     self.param_dtype = param_dtype
     self.carry_init = carry_init
-    self.rngs = rngs
+    self.rngs: rnglib.RngStream | None
+    if keep_rngs:
+      self.rngs = rngs.carry.fork()
+    else:
+      self.rngs = None
 
     # Combine input transformations into a single linear layer
     self.dense_i = Linear(
@@ -568,7 +616,9 @@ class GRUCell(RNNCellBase):
     return new_h, new_h
 
   def initialize_carry(
-    self, input_shape: tuple[int, ...], rngs: rnglib.Rngs | None = None
+    self,
+    input_shape: tuple[int, ...],
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
   ) -> Array:  # type: ignore[override]
     """Initialize the RNN cell carry.
 
@@ -582,8 +632,13 @@ class GRUCell(RNNCellBase):
     batch_dims = input_shape[:-1]
     if rngs is None:
       rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry
+    if rngs is None:
+      raise ValueError('RNGs must be provided to initialize the cell carry.')
+
     mem_shape = batch_dims + (self.hidden_features,)
-    h = self.carry_init(rngs.carry(), mem_shape, self.param_dtype)
+    h = self.carry_init(rngs(), mem_shape, self.param_dtype)
     return h
 
   @property
@@ -602,14 +657,15 @@ class RNN(Module):
   def __init__(
     self,
     cell: RNNCellBase,
+    *,
     time_major: bool = False,
     return_carry: bool = False,
     reverse: bool = False,
     keep_order: bool = False,
     unroll: int = 1,
-    rngs: rnglib.Rngs | None = None,
     state_axes: Mapping[str, int | type[iteration.Carry] | None] | None = None,
     broadcast_rngs: filterlib.Filter = None,
+    rngs: rnglib.Rngs | rnglib.RngStream | bool = True,
   ):
     self.cell = cell
     self.time_major = time_major
@@ -617,9 +673,18 @@ class RNN(Module):
     self.reverse = reverse
     self.keep_order = keep_order
     self.unroll = unroll
-    if rngs is None:
-      rngs = rnglib.Rngs(0)
-    self.rngs = rngs
+    self.rngs: rnglib.RngStream | None
+    if rngs is True:
+      self.rngs = rnglib.RngStream(0, tag='carry')
+    elif isinstance(rngs, rnglib.Rngs):
+      self.rngs = rngs.carry.fork()
+    elif rngs is False:
+      self.rngs = None
+    else:
+      raise ValueError(
+        'Expected rngs to be a jax.Array, int, Rngs, or bool. '
+        f'Got {type(rngs)}.'
+      )
     self.state_axes = state_axes or {...: iteration.Carry}  # type: ignore
     self.broadcast_rngs = broadcast_rngs
 
@@ -633,7 +698,7 @@ class RNN(Module):
     time_major: bool | None = None,
     reverse: bool | None = None,
     keep_order: bool | None = None,
-    rngs: rnglib.Rngs | None = None,
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
   ):
     if return_carry is None:
       return_carry = self.return_carry
@@ -661,23 +726,26 @@ class RNN(Module):
     # maybe reverse the sequence
     if reverse:
       inputs = jax.tree_util.tree_map(
-                lambda x: flip_sequences(
-                    x,
-                    seq_lengths,
-                    num_batch_dims=len(batch_dims),
-                    time_major=time_major,  # type: ignore
-                ),
-                inputs,
-            )
+        lambda x: flip_sequences(
+          x,
+          seq_lengths,
+          num_batch_dims=len(batch_dims),
+          time_major=time_major,  # type: ignore
+        ),
+        inputs,
+      )
     if rngs is None:
       rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry.fork()
+
     carry: Carry = (
-            self.cell.initialize_carry(
-                inputs.shape[:time_axis] + inputs.shape[time_axis + 1 :], rngs
-            )
-            if initial_carry is None
-            else initial_carry
-        )
+      self.cell.initialize_carry(
+        inputs.shape[:time_axis] + inputs.shape[time_axis + 1 :], rngs
+      )
+      if initial_carry is None
+      else initial_carry
+    )
 
     slice_carry = seq_lengths is not None and return_carry
     broadcast_rngs = nnx.All(nnx.RngState, self.broadcast_rngs)
@@ -812,131 +880,144 @@ def _concatenate(a: Array, b: Array) -> Array:
 
 
 class RNNBase(Protocol):
-    def __call__(
-        self,
-        inputs: Array,
-        *,
-        initial_carry: Carry | None = None,
-        rngs: rnglib.Rngs | None = None,
-        seq_lengths: Array | None = None,
-        return_carry: bool | None = None,
-        time_major: bool | None = None,
-        reverse: bool | None = None,
-        keep_order: bool | None = None,
-    ) -> Output | tuple[Carry, Output]: ...
+  def __call__(
+    self,
+    inputs: Array,
+    *,
+    initial_carry: Carry | None = None,
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
+    seq_lengths: Array | None = None,
+    return_carry: bool | None = None,
+    time_major: bool | None = None,
+    reverse: bool | None = None,
+    keep_order: bool | None = None,
+  ) -> Output | tuple[Carry, Output]: ...
 
 
 class Bidirectional(Module):
-    """Processes the input in both directions and merges the results.
+  """Processes the input in both directions and merges the results.
 
-    Example usage::
+  Example usage::
 
-      >>> from flax import nnx
-      >>> import jax
-      >>> import jax.numpy as jnp
+    >>> from flax import nnx
+    >>> import jax
+    >>> import jax.numpy as jnp
 
-      >>> # Define forward and backward RNNs
-      >>> forward_rnn = RNN(GRUCell(in_features=3, hidden_features=4, rngs=nnx.Rngs(0)))
-      >>> backward_rnn = RNN(GRUCell(in_features=3, hidden_features=4, rngs=nnx.Rngs(0)))
+    >>> # Define forward and backward RNNs
+    >>> forward_rnn = RNN(GRUCell(in_features=3, hidden_features=4, rngs=nnx.Rngs(0)))
+    >>> backward_rnn = RNN(GRUCell(in_features=3, hidden_features=4, rngs=nnx.Rngs(0)))
 
-      >>> # Create Bidirectional layer
-      >>> layer = Bidirectional(forward_rnn=forward_rnn, backward_rnn=backward_rnn)
+    >>> # Create Bidirectional layer
+    >>> layer = Bidirectional(forward_rnn=forward_rnn, backward_rnn=backward_rnn)
 
-      >>> # Input data
-      >>> x = jnp.ones((2, 3, 3))
+    >>> # Input data
+    >>> x = jnp.ones((2, 3, 3))
 
-      >>> # Apply the layer
-      >>> out = layer(x)
-      >>> print(out.shape)
-      (2, 3, 8)
+    >>> # Apply the layer
+    >>> out = layer(x)
+    >>> print(out.shape)
+    (2, 3, 8)
 
-    """
+  """
 
-    forward_rnn: RNNBase
-    backward_rnn: RNNBase
-    merge_fn: Callable[[Array, Array], Array] = _concatenate
-    time_major: bool = False
-    return_carry: bool = False
+  forward_rnn: RNNBase
+  backward_rnn: RNNBase
+  merge_fn: Callable[[Array, Array], Array] = _concatenate
+  time_major: bool = False
+  return_carry: bool = False
 
-    def __init__(
-        self,
-        forward_rnn: RNNBase,
-        backward_rnn: RNNBase,
-        *,
-        merge_fn: Callable[[Array, Array], Array] = _concatenate,
-        time_major: bool = False,
-        return_carry: bool = False,
-        rngs: rnglib.Rngs | None = None,
-    ):
-        self.forward_rnn = forward_rnn
-        self.backward_rnn = backward_rnn
-        self.merge_fn = merge_fn
-        self.time_major = time_major
-        self.return_carry = return_carry
-        if rngs is None:
-            rngs = rnglib.Rngs(0)
-        self.rngs = rngs
+  def __init__(
+    self,
+    forward_rnn: RNNBase,
+    backward_rnn: RNNBase,
+    *,
+    merge_fn: Callable[[Array, Array], Array] = _concatenate,
+    time_major: bool = False,
+    return_carry: bool = False,
+    rngs: rnglib.Rngs | rnglib.RngStream | bool = True,
+  ):
+    self.forward_rnn = forward_rnn
+    self.backward_rnn = backward_rnn
+    self.merge_fn = merge_fn
+    self.time_major = time_major
+    self.return_carry = return_carry
+    self.rngs: rnglib.RngStream | None
+    if rngs is True:
+      self.rngs = rnglib.RngStream(0, tag='carry')
+    elif rngs is False:
+      self.rngs = None
+    elif isinstance(rngs, rnglib.Rngs):
+      self.rngs = rngs.carry.fork()
+    elif isinstance(rngs, rnglib.RngStream):
+      self.rngs = rngs
+    else:
+      raise TypeError(
+        f'rngs must be a Rngs, jax.Array, int, or bool, but got {type(rngs)}.'
+      )
 
-    def __call__(
-        self,
-        inputs: Array,
-        *,
-        initial_carry: tuple[Carry, Carry] | None = None,
-        rngs: rnglib.Rngs | None = None,
-        seq_lengths: Array | None = None,
-        return_carry: bool | None = None,
-        time_major: bool | None = None,
-        reverse: bool | None = None,  # unused
-        keep_order: bool | None = None,  # unused
-    ) -> Output | tuple[tuple[Carry, Carry], Output]:
-        if time_major is None:
-            time_major = self.time_major
-        if return_carry is None:
-            return_carry = self.return_carry
-        if rngs is None:
-            rngs = self.rngs
-        if initial_carry is not None:
-            initial_carry_forward, initial_carry_backward = initial_carry
-        else:
-            initial_carry_forward = None
-            initial_carry_backward = None
-        # Throw a warning in case the user accidentally re-uses the forward RNN
-        # for the backward pass and does not intend for them to share parameters.
-        if self.forward_rnn is self.backward_rnn:
-            logging.warning(
-                "forward_rnn and backward_rnn is the same object, so "
-                "they will share parameters."
-            )
+  def __call__(
+    self,
+    inputs: Array,
+    *,
+    initial_carry: tuple[Carry, Carry] | None = None,
+    rngs: rnglib.Rngs | rnglib.RngStream | None = None,
+    seq_lengths: Array | None = None,
+    return_carry: bool | None = None,
+    time_major: bool | None = None,
+    reverse: bool | None = None,  # unused
+    keep_order: bool | None = None,  # unused
+  ) -> Output | tuple[tuple[Carry, Carry], Output]:
+    if time_major is None:
+      time_major = self.time_major
+    if return_carry is None:
+      return_carry = self.return_carry
+    if rngs is None:
+      rngs = self.rngs
+    if isinstance(rngs, rnglib.Rngs):
+      rngs = rngs.carry
 
-        # Encode in the forward direction.
-        carry_forward, outputs_forward = self.forward_rnn(
-            inputs,
-            initial_carry=initial_carry_forward,
-            rngs=rngs,
-            seq_lengths=seq_lengths,
-            return_carry=True,
-            time_major=time_major,
-            reverse=False,
-        )
+    if initial_carry is not None:
+      initial_carry_forward, initial_carry_backward = initial_carry
+    else:
+      initial_carry_forward = None
+      initial_carry_backward = None
+    # Throw a warning in case the user accidentally re-uses the forward RNN
+    # for the backward pass and does not intend for them to share parameters.
+    if self.forward_rnn is self.backward_rnn:
+      logging.warning(
+        'forward_rnn and backward_rnn is the same object, so '
+        'they will share parameters.'
+      )
 
-        # Encode in the backward direction.
-        carry_backward, outputs_backward = self.backward_rnn(
-            inputs,
-            initial_carry=initial_carry_backward,
-            rngs=rngs,
-            seq_lengths=seq_lengths,
-            return_carry=True,
-            time_major=time_major,
-            reverse=True,
-            keep_order=True,
-        )
+    # Encode in the forward direction.
+    carry_forward, outputs_forward = self.forward_rnn(
+      inputs,
+      initial_carry=initial_carry_forward,
+      rngs=rngs,
+      seq_lengths=seq_lengths,
+      return_carry=True,
+      time_major=time_major,
+      reverse=False,
+    )
 
-        carry = (carry_forward, carry_backward) if return_carry else None
-        outputs = jax.tree_util.tree_map(
-            self.merge_fn, outputs_forward, outputs_backward
-        )
+    # Encode in the backward direction.
+    carry_backward, outputs_backward = self.backward_rnn(
+      inputs,
+      initial_carry=initial_carry_backward,
+      rngs=rngs,
+      seq_lengths=seq_lengths,
+      return_carry=True,
+      time_major=time_major,
+      reverse=True,
+      keep_order=True,
+    )
 
-        if return_carry:
-            return carry, outputs
-        else:
-            return outputs
+    carry = (carry_forward, carry_backward) if return_carry else None
+    outputs = jax.tree_util.tree_map(
+      self.merge_fn, outputs_forward, outputs_backward
+    )
+
+    if return_carry:
+      return carry, outputs
+    else:
+      return outputs
