@@ -43,7 +43,7 @@ N = tp.TypeVar('N', bound=Module)
 T = tp.TypeVar('T')
 StrInt = tp.TypeVar('StrInt', str, int)
 AxisName = tp.Hashable
-Leaves = tp.List[Leaf]
+Leaves = list[Leaf]
 Index = int
 
 
@@ -112,7 +112,10 @@ class StateAxes(extract.PrefixMapping):
     return hash((self.filters, self.axes))
 
 
-AxisFn = tp.Callable[[graph.GraphState, int, tp.Mapping], graph.GraphState]
+AxisFn = tp.Callable[
+  [graph.GraphState | variablelib.Variable, int, tp.Mapping],
+  graph.GraphState | variablelib.Variable,
+]
 
 
 def _update_variable_sharding_metadata(
@@ -124,13 +127,13 @@ def _update_variable_sharding_metadata(
     ):
       if isinstance(node_states.metadata, int):
         state = node_states.state
-        assert isinstance(state, State)
+        assert isinstance(state, State | variablelib.Variable)
         state = axis_fn(state, node_states.metadata, transform_metadata)
         return node_states.replace(states=(state,))
       else:
-        states_out: list[graph.GraphState] = []
+        states_out: list[graph.GraphState | variablelib.Variable] = []
         for state, axis in zip(node_states.states, node_states.metadata.axes):
-          assert isinstance(state, graph.State)
+          assert isinstance(state, graph.State | variablelib.Variable)
           if isinstance(axis, int):
             state = axis_fn(state, axis, transform_metadata)
           states_out.append(state)
@@ -165,7 +168,7 @@ class VmapFn:
       pure_args = _update_variable_sharding_metadata(
           pure_args, self.transform_metadata, spmd.remove_axis
       )
-    args = extract.from_tree(pure_args, ctxtag='vmap')
+    args = extract.from_tree(pure_args, ctxtag='vmap', is_inner=True)
 
     out = self.f(*args)
 
@@ -233,7 +236,7 @@ def vmap(
       addition to integers and None, :class:`StateAxes`  can be used to control
       how graph nodes like Modules are vectorized by specifying the axes to be
       applied to substates of the graph node given a `Filter
-      <https://flax.readthedocs.io/en/latest/nnx/filters_guide.html>`__.
+      <https://flax.readthedocs.io/en/latest/guides/filters_guide.html>`__.
     out_axes: An integer, None, or pytree indicating where the mapped axis
       should appear in the output (see `jax.vmap
       <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`__).
@@ -343,7 +346,9 @@ def vmap(
         args, prefix=in_axes, split_fn=_vmap_split_fn, ctxtag='vmap'
     )
     pure_args_out, pure_out = vmapped_fn(*pure_args)
-    _args_out, out = extract.from_tree((pure_args_out, pure_out), ctxtag='vmap')
+    _args_out, out = extract.from_tree(
+      (pure_args_out, pure_out), ctxtag='vmap', is_inner=False
+    )
     return out
 
   return vmap_wrapper  # type: ignore
@@ -369,7 +374,7 @@ class PmapFn:
       pure_args = _update_variable_sharding_metadata(
           pure_args, self.transform_metadata, spmd.remove_axis
       )
-    args = extract.from_tree(pure_args, ctxtag='pmap')
+    args = extract.from_tree(pure_args, ctxtag='pmap', is_inner=True)
 
     out = self.f(*args)
 
@@ -449,7 +454,7 @@ def pmap(
       addition to integers and None, :class:`StateAxes`  can be used to control
       how graph nodes like Modules are vectorized by specifying the axes to be
       applied to substates of the graph node given a `Filter
-      <https://flax.readthedocs.io/en/latest/nnx/filters_guide.html>`__.
+      <https://flax.readthedocs.io/en/latest/guides/filters_guide.html>`__.
     out_axes: An integer, None, or pytree indicating where the mapped axis
       should appear in the output (see `jax.vmap
       <https://jax.readthedocs.io/en/latest/_autosummary/jax.vmap.html>`__).
@@ -566,7 +571,9 @@ def pmap(
         args, prefix=in_axes, split_fn=_vmap_split_fn, ctxtag='pmap'
     )
     pure_args_out, pure_out = pmapped_fn(*pure_args)
-    _args_out, out = extract.from_tree((pure_args_out, pure_out), ctxtag='pmap')
+    _args_out, out = extract.from_tree(
+      (pure_args_out, pure_out), ctxtag='pmap', is_inner=False
+    )
     return out
 
   return vmap_wrapper  # type: ignore
@@ -645,24 +652,23 @@ def _check_carry_same_references(carry_arg, carry_arg_out):
       )
 
   jax.tree_util.tree_map_with_path(
-    check_carry_same_references, carry_arg, carry_arg_out
+      check_carry_same_references,
+      carry_arg,
+      carry_arg_out,
+      is_leaf=lambda x: graph.is_graph_node(x)
+      and not isinstance(x, variablelib.Variable),
   )
 
-def _extract_index_mappings(
-  pure_carry_arg_out,
-  carry_index_mappings: list[graph.HashableMapping[int, int]],
-  /,
+def _extract_graphdefs(
+  pure_carry_arg_out, carry_graphdefs: list[graph.GraphDef], /
 ):
   def extract_index_mappings(x):
     if isinstance(x, extract.NodeStates) and isinstance(
-      x._graphdef, graph.NodeDef
+      x._graphdef, graph.GraphDef
     ):
-      index_mapping = x._graphdef.index_mapping
-      assert index_mapping is not None
-      carry_index_mappings.append(index_mapping)
-      x = x.replace(
-        _graphdef=dataclasses.replace(x._graphdef, index_mapping=None)
-      )
+      graphdef = x._graphdef
+      carry_graphdefs.append(graphdef)
+      x = x.replace(_graphdef=graphdef.with_no_outer_index())
     return x
 
   pure_carry_arg_out = jax.tree.map(
@@ -673,19 +679,17 @@ def _extract_index_mappings(
 
   return pure_carry_arg_out
 
-def _insert_index_mappings(
+def _insert_graphdefs(
   pure_carry_arg_out,
-  carry_index_mappings: deque[graph.HashableMapping[int, int]],
+  carry_graphdefs: deque[graph.GraphDef],
   /,
 ):
   def insert_index_mappings(x):
     if isinstance(x, extract.NodeStates) and isinstance(
-      x._graphdef, graph.NodeDef
+      x._graphdef, graph.GraphDef
     ):
-      index_mapping = carry_index_mappings.popleft()
-      x = x.replace(
-        _graphdef=dataclasses.replace(x._graphdef, index_mapping=index_mapping)
-      )
+      graphdef = carry_graphdefs.popleft()
+      x = x.replace(_graphdef=graphdef)
     return x
 
   pure_carry_arg_out = jax.tree.map(
@@ -697,19 +701,19 @@ def _insert_index_mappings(
 
 
 def _scan_split_in(
-    carry_deque: PytreeDeque[list[State]],
-    broadcast_deque: PytreeDeque[list[State]],
-    broadcast_arrays: PytreeDeque[Broadcasted],
-    /,
-    ctx: graph.SplitContext,
-    path,
-    prefix,
-    x,
+  carry_deque: PytreeDeque[list[State | variablelib.Variable]],
+  broadcast_deque: PytreeDeque[list[State | variablelib.Variable]],
+  broadcast_arrays: PytreeDeque[Broadcasted],
+  /,
+  ctx: graph.SplitContext,
+  path,
+  prefix,
+  x,
 ):
-  if graph.is_graph_node(x):
-    vectorized_states: list[State] = []
-    carry_states: list[State] = []
-    broadcast_states: list[State] = []
+  if graph.is_graph_node(x) or isinstance(x, variablelib.Variable):
+    vectorized_states: list[State | variablelib.Variable] = []
+    carry_states: list[State | variablelib.Variable] = []
+    broadcast_states: list[State | variablelib.Variable] = []
     if isinstance(prefix, StateAxes):
       graphdef, *states = ctx.split(x, *prefix.filters)
 
@@ -778,21 +782,21 @@ def _scan_split_in(
 
 
 def _scan_split_out(
-    carry_deque: PytreeDeque[list[State]],
-    broadcast_deque: PytreeDeque[list[State]],
-    /,
-    ctx: graph.SplitContext,
-    path: extract.KeyPath,
-    prefix,
-    x,
+  carry_deque: PytreeDeque[list[State | variablelib.Variable]],
+  broadcast_deque: PytreeDeque[list[State | variablelib.Variable]],
+  /,
+  ctx: graph.SplitContext,
+  path: extract.KeyPath,
+  prefix,
+  x,
 ):
   assert isinstance(path[0], jax.tree_util.SequenceKey)
   is_input_arg = path[0].idx == 0
 
-  if graph.is_graph_node(x):
-    vectorized_states: list[State] = []
-    carry_states: list[State] = []
-    broadcast_states: list[State] = []
+  if graph.is_graph_node(x) or isinstance(x, variablelib.Variable):
+    vectorized_states: list[State | variablelib.Variable] = []
+    carry_states: list[State | variablelib.Variable] = []
+    broadcast_states: list[State | variablelib.Variable] = []
     if isinstance(prefix, StateAxes):
       graphdef, *states = ctx.split(x, *prefix.filters)
 
@@ -868,14 +872,14 @@ def _scan_split_out(
 
 
 def _scan_merge_in(
-    carry_deque: PytreeDeque[list[State]],
-    broadcast_deque: PytreeDeque[list[State]],
-    broadcast_arrays: PytreeDeque[Broadcasted],
-    /,
-    ctx: graph.MergeContext,
-    path,
-    prefix,
-    x,
+  carry_deque: PytreeDeque[list[State]],
+  broadcast_deque: PytreeDeque[list[State]],
+  broadcast_arrays: PytreeDeque[Broadcasted],
+  /,
+  ctx: graph.MergeContext,
+  path,
+  prefix,
+  x,
 ):
   if isinstance(x, extract.NodeStates):
     carry_states = carry_deque.popleft()
@@ -889,13 +893,13 @@ def _scan_merge_in(
 
 
 def _scan_merge_out(
-    carry_deque: PytreeDeque[list[State]],
-    broadcast_deque: PytreeDeque[list[State]],
-    /,
-    ctx: graph.MergeContext,
-    path,
-    prefix,
-    x,
+  carry_deque: PytreeDeque[list[State]],
+  broadcast_deque: PytreeDeque[list[State]],
+  /,
+  ctx: graph.MergeContext,
+  path,
+  prefix,
+  x,
 ):
   assert isinstance(path[0], jax.tree_util.SequenceKey)
   is_input_arg = path[0].idx == 0
@@ -1017,6 +1021,7 @@ class ScanFn:
       is_leaf=lambda x: isinstance(x, (extract.NodeStates, Broadcasted)),
       map_non_graph_nodes=True,
       ctxtag='scan',
+      is_inner=True,
     )
     assert not carry_deque and not broadcast_deque and not broadcast_arrays
 
@@ -1062,8 +1067,10 @@ class ScanFn:
       assert self.input_carry_argnum is None
       assert carry_arg_out is None
 
-    carry_deque_out = PytreeDeque[list[State]]()
-    _broadcast_deque_out_tmp = PytreeDeque[list[State]]()  # discarded
+    carry_deque_out = PytreeDeque[list[State | variablelib.Variable]]()
+    _broadcast_deque_out_tmp = PytreeDeque[
+      list[State | variablelib.Variable]
+    ]()  # discarded
     pure_args_out: tuple
     pure_args_out, pure_out = extract.to_tree(
       (args_out, out),
@@ -1094,12 +1101,10 @@ class ScanFn:
       assert self.input_carry_argnum is None
       pure_carry_arg_out = None
 
-    # next we have to remove all the index_mappings from the NodeDefs
+    # next we have to remove all the index_mappings from the GraphDefs
     # in the carry outputs because they are not present in the inputs
-    carry_index_mappings: list[graph.HashableMapping[int, int]] = []
-    pure_carry_arg_out = _extract_index_mappings(
-      pure_carry_arg_out, carry_index_mappings
-    )
+    carry_graphdefs: list[graph.GraphDef] = []
+    pure_carry_arg_out = _extract_graphdefs(pure_carry_arg_out, carry_graphdefs)
 
     carry_arg_out = (
       pure_carry_arg_out,
@@ -1108,7 +1113,7 @@ class ScanFn:
       broadcast_arrays_out,
     )
     scan_out = (
-      graph.Static(tuple(carry_index_mappings)),
+      carry_graphdefs,
       pure_args_out,
       pure_out,
     )
@@ -1248,16 +1253,15 @@ def scan(
         broadcast_arrays_out,
     ) = carry_out
     (
-      static_carry_index_mappings,
+      carry_graphdefs,
       pure_args_out,
       pure_out,
     ) = scan_out
 
-    # next we have to insert all the index_mappings back into the NodeDefs
+    # next we have to insert all the index_mappings back into the GraphDefs
     # in the carry outputs
-    carry_index_mappings = deque(static_carry_index_mappings.value)
-    pure_carry_arg_out = _insert_index_mappings(
-      pure_carry_arg_out, carry_index_mappings
+    pure_carry_arg_out = _insert_graphdefs(
+      pure_carry_arg_out, deque(carry_graphdefs)
     )
 
     # insert pure carry into pure_args_out
@@ -1280,6 +1284,7 @@ def scan(
       is_leaf=lambda x: isinstance(x, (extract.NodeStates, Broadcasted)),
       map_non_graph_nodes=True,
       ctxtag='scan',
+      is_inner=False,
     )
 
     # extract the carry from args_out
@@ -1307,9 +1312,6 @@ def scan(
   return scan_wrapper  # type: ignore
 
 
-
-
-
 # -------------------------------
 # while_loop
 # -------------------------------
@@ -1329,36 +1331,14 @@ class WhileLoopCondFn:
 
 
 def _add_fake_index_mapping(tree: tp.Any):
-  global_index_mapping = {}  # for the whole context, over all inputs
-  def per_node_state(ns: extract.NodeStates | tp.Any):
-    if not isinstance(ns, extract.NodeStates) or not isinstance(
-      ns._graphdef, graph.NodeDef
+  def per_node_state(node_state: extract.NodeStates | tp.Any):
+    if not isinstance(node_state, extract.NodeStates) or not isinstance(
+      node_state._graphdef, graph.GraphDef
     ):
-      return ns
+      return node_state
 
-    def per_node_def(nd: graph.NodeDef | graph.NodeRef):
-      if nd.index >= 0:
-        global_index_mapping[nd.index] = nd.index
-      if isinstance(nd, graph.NodeRef):
-        return
-
-      for attribute in nd.attributes:
-        if type(attribute) is graph.SubGraphAttribute:
-          per_node_def(attribute.value)
-        elif (
-          type(attribute) is graph.LeafAttribute
-          and isinstance(attribute.value, (graph.VariableDef, graph.NodeRef))
-          and attribute.value.index >= 0
-        ):
-          global_index_mapping[attribute.value.index] = attribute.value.index
-      return
-
-    per_node_def(ns._graphdef)
     return dataclasses.replace(
-      ns,
-      _graphdef=dataclasses.replace(
-        ns._graphdef, index_mapping=graph.HashableMapping(global_index_mapping)
-      ),
+      node_state, _graphdef=node_state._graphdef.with_same_outer_index()
     )
 
   return jax.tree.map(per_node_state, tree,
@@ -1366,16 +1346,18 @@ def _add_fake_index_mapping(tree: tp.Any):
 
 
 def _remove_index_mapping(tree: tp.Any):
-  '''Remove a fake index_mapping for the input to match that of the output.'''
-  def per_node_state(ns: extract.NodeStates | tp.Any):
-    if not isinstance(ns, extract.NodeStates) or not isinstance(
-      ns._graphdef, graph.NodeDef
+  """Remove a fake outer_index for the input to match that of the output."""
+
+  def per_node_state(node_state: extract.NodeStates | tp.Any):
+    if not isinstance(node_state, extract.NodeStates) or not isinstance(
+      node_state._graphdef, graph.GraphDef
     ):
-      return ns
-    assert isinstance(ns._graphdef, graph.NodeDef)
-    return dataclasses.replace(ns, _graphdef=dataclasses.replace(
-      ns._graphdef, index_mapping=None
-    ))
+      return node_state
+    assert isinstance(node_state._graphdef, graph.GraphDef)
+    node_state = dataclasses.replace(
+      node_state, _graphdef=node_state._graphdef.with_no_outer_index()
+    )
+    return node_state
 
   return jax.tree.map(per_node_state, tree,
                       is_leaf=lambda x: isinstance(x, extract.NodeStates))
@@ -1393,19 +1375,23 @@ class WhileLoopBodyFn:
     # Removing the dummy index mapping being added outside of body function.
     pure_val_in = _remove_index_mapping(pure_val)
 
-    val = extract.from_tree(pure_val_in, ctxtag='while_loop_body')
+    val = extract.from_tree(
+      pure_val_in, ctxtag='while_loop_body', is_inner=True
+    )
     out = self.f(val)
     pure_out = extract.to_tree(out, ctxtag='while_loop_body')
 
     try:
       jax.tree.map(lambda a, b: None, pure_val, pure_out)
     except ValueError as e:
-      msg = ("nnx.while_loop requires body function's input and output to "
-             "have the same reference and pytree structure, but they differ. "
-             "If the mismatch comes from `index_mapping` field, you might "
-             "have modified reference structure within the body function, "
-             "which is not allowed."
-             f"Detail of the mismatch: \n {str(e)}")
+      msg = (
+        "nnx.while_loop requires body function's input and output to "
+        'have the same reference and pytree structure, but they differ. '
+        'If the mismatch comes from `outer_index` field, you might '
+        'have modified reference structure within the body function, '
+        'which is not allowed.'
+        f'Detail of the mismatch: \n {str(e)}'
+      )
       raise ValueError(msg)
 
     return pure_out
@@ -1456,7 +1442,7 @@ def while_loop(cond_fun: tp.Callable[[T], tp.Any],
     WhileLoopBodyFn(body_fun),
     pure_init_val,
   )
-  out = extract.from_tree(pure_out, ctxtag='while_loop')
+  out = extract.from_tree(pure_out, ctxtag='while_loop', is_inner=False)
   return out
 
 
@@ -1472,19 +1458,21 @@ class ForiLoopBodyFn:
     # Removing the dummy index mapping being added outside of body function.
     pure_val_in = _remove_index_mapping(pure_val)
 
-    val = extract.from_tree(pure_val_in, ctxtag='fori_loop_body')
+    val = extract.from_tree(pure_val_in, ctxtag='fori_loop_body', is_inner=True)
     out = self.f(i, val)
     pure_out = extract.to_tree(out, ctxtag='fori_loop_body')
 
     try:
       jax.tree.map(lambda a, b: None, pure_val, pure_out)
     except ValueError as e:
-      msg = ("nnx.fori_loop requires body function's input and output to "
-             "have the same reference and pytree structure, but they differ. "
-             "If the mismatch comes from `index_mapping` field, you might "
-             "have modified reference structure within the body function, "
-             "which is not allowed. "
-             f"Detail of the mismatch: \n {str(e)}")
+      msg = (
+        "nnx.fori_loop requires body function's input and output to "
+        'have the same reference and pytree structure, but they differ. '
+        'If the mismatch comes from `outer_index` field, you might '
+        'have modified reference structure within the body function, '
+        'which is not allowed. '
+        f'Detail of the mismatch: \n {str(e)}'
+      )
       raise ValueError(msg)
 
     return pure_out
@@ -1545,5 +1533,5 @@ def fori_loop(lower: int, upper: int,
   pure_out = jax.lax.fori_loop(lower, upper,
                                ForiLoopBodyFn(body_fun), pure_init_val,
                                unroll=unroll)
-  out = extract.from_tree(pure_out, ctxtag='fori_loop')
+  out = extract.from_tree(pure_out, ctxtag='fori_loop', is_inner=False)
   return out
